#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-2-Clause
# Copyright 2018 Linaro Ltd.
# Copyright 2018 Arm Ltd.

import sys
import argparse
import json

import ruamel.yaml
import dtschema


def main():
    ap = argparse.ArgumentParser(fromfile_prefix_chars='@',
        epilog='Arguments can also be passed in a file prefixed with a "@" character.')
    ap.add_argument("-o", "--outfile", type=str,
                    help="Filename of the processed schema")
    ap.add_argument("-j", "--json", help="Encode the processed schema in json",
                    action="store_true")
    ap.add_argument("schemas", nargs='*', type=str,
                    help="Names of directories, or YAML encoded schema files")
    ap.add_argument('-u', '--useronly', help="Only process user schemas", action="store_true")
    ap.add_argument('-V', '--version', help="Print version number",
                    action="version", version=dtschema.__version__)
    args = ap.parse_args()

    schemas = dtschema.DTValidator(args.schemas).schemas
    if not schemas:
        return -1

    if args.outfile:
        f = open(args.outfile, 'w', encoding='utf-8')
    else:
        f = sys.stdout

    if (args.json):
        json.dump(schemas, f, indent=4)
    else:
        yaml = ruamel.yaml.YAML(typ='safe')
        yaml.dump(schemas, f)



# SPDX-License-Identifier: BSD-2-Clause
# Copyright 2018 Linaro Ltd.
# Copyright 2018-2023 Arm Ltd.
# Python library for Devicetree schema validation
import sys
import os
import re
import copy
import jsonschema

import dtschema

from jsonschema.exceptions import RefResolutionError

schema_base_url = "http://devicetree.org/"
schema_basedir = os.path.dirname(os.path.abspath(__file__))


def path_to_obj(tree, path):
    for pc in path:
        tree = tree[pc]
    return tree


def get_line_col(tree, path, obj=None):
    import ruamel.yaml

    if isinstance(obj, ruamel.yaml.comments.CommentedBase):
        return obj.lc.line, obj.lc.col
    obj = path_to_obj(tree, path)
    if isinstance(obj, ruamel.yaml.comments.CommentedBase):
        return obj.lc.line, obj.lc.col
    if len(path) < 1:
        return -1, -1
    obj = path_to_obj(tree, list(path)[:-1])
    if isinstance(obj, ruamel.yaml.comments.CommentedBase):
        if path[-1] == '$nodename':
            return -1, -1
        return obj.lc.key(path[-1])
    return -1, -1

def _is_node_schema(schema):
    return isinstance(schema, dict) and \
           (('type' in schema and schema['type'] == 'object') or
            schema.keys() & {'properties', 'patternProperties'})


def _schema_allows_no_undefined_props(schema):
    return not schema.get("additionalProperties", True) or \
           not schema.get("unevaluatedProperties", True)

class DTSchema(dict):
    DtValidator = jsonschema.validators.extend(
        jsonschema.Draft201909Validator,
        format_checker=jsonschema.FormatChecker(),
        version='DT')

    def __init__(self, schema_file, line_numbers=False):
        self.paths = [(schema_base_url, schema_basedir + '/')]
        with open(schema_file, 'r', encoding='utf-8') as f:
            import ruamel.yaml

            if line_numbers:
                rtyaml = ruamel.yaml.YAML(typ='rt')
                rtyaml.allow_duplicate_keys = False
                rtyaml.preserve_quotes = True

                schema = rtyaml.load(f.read())
            else:
                yaml = ruamel.yaml.YAML(typ='safe')
                yaml.allow_duplicate_keys = False
                schema = yaml.load(f.read())

        self.filename = os.path.abspath(schema_file)

        id = schema['$id'].rstrip('#')
        match = re.search('(.*/schemas/)(.+)$', id)
        self.base_path = os.path.abspath(schema_file)[:-len(match[2])]

        super().__init__(schema)

    def http_handler(self, uri):
        '''Custom handler for http://devicetree.org references'''
        uri = uri.rstrip('#')
        for p in self.paths:
            filename = uri.replace(p[0], p[1])
            if not os.path.isfile(filename):
                continue
            with open(filename, 'r', encoding='utf-8') as f:
                import ruamel.yaml
                yaml = ruamel.yaml.YAML(typ='safe')
                yaml.allow_duplicate_keys = False
                return yaml.load(f.read())

        raise RefResolutionError('Error in referenced schema matching $id: ' + uri)

    def annotate_error(self, error, schema, path):
        error.note = None
        error.schema_file = None

        for e in error.context:
            self.annotate_error(e, schema, path + e.schema_path)

        scope = self.validator.ID_OF(schema)
        self.validator.resolver.push_scope(scope)
        ref_depth = 1

        for p in path:
            while p not in schema and '$ref' in schema and isinstance(schema['$ref'], str):
                ref = self.validator.resolver.resolve(schema['$ref'])
                schema = ref[1]
                self.validator.resolver.push_scope(ref[0])
                ref_depth += 1

            if '$id' in schema and isinstance(schema['$id'], str):
                error.schema_file = schema['$id']

            schema = schema[p]

            if isinstance(schema, dict):
                if 'description' in schema and isinstance(schema['description'], str):
                    error.note = schema['description']

        while ref_depth > 0:
            self.validator.resolver.pop_scope()
            ref_depth -= 1

        if isinstance(error.schema, dict) and 'description' in error.schema:
            error.note = error.schema['description']

    def iter_errors(self):
        self.resolver = jsonschema.RefResolver.from_schema(self,
                            handlers={'http': self.http_handler})
        meta_schema = self.resolver.resolve_from_url(self['$schema'])
        self.validator = self.DtValidator(meta_schema, resolver=self.resolver)

        for error in self.validator.iter_errors(self):
            scherr = jsonschema.exceptions.SchemaError.create_from(error)
            self.annotate_error(scherr, meta_schema, scherr.schema_path)
            scherr.linecol = get_line_col(self, scherr.path)
            yield scherr

    def is_valid(self, strict=False):
        ''' Check if schema passes validation against json-schema.org schema '''
        if strict:
            for error in self.iter_errors():
                raise error
        else:
            # Using the draft7 metaschema because 2019-09 with $recursiveRef seems broken
            # Probably fixed with referencing library
            for error in self.DtValidator(jsonschema.Draft7Validator.META_SCHEMA).iter_errors(self):
                scherr = jsonschema.exceptions.SchemaError.create_from(error)
                raise scherr

    def fixup(self):
        processed_schema = copy.deepcopy(dict(self))
        dtschema.fixups.fixup_schema(processed_schema)
        return processed_schema

    def _check_schema_refs(self, schema, parent=None, is_common=False, has_constraint=False):
        if not parent:
            is_common = not _schema_allows_no_undefined_props(schema)
        if isinstance(schema, dict):
            if parent in {'if', 'select', 'definitions', '$defs', 'then',
                          'else', 'dependencies', 'dependentSchemas'}:
                return

            if _is_node_schema(schema):
                has_constraint = _schema_allows_no_undefined_props(schema)

            if not is_common and _is_node_schema(schema) and \
               (schema.keys() & {'properties', 'patternProperties', '$ref'}):

                ref_has_constraint = False
                if 'properties' in schema:
                    # All false properties are not complete schemas
                    for prop,val in schema['properties'].items():
                        if val is not False:
                            break
                    else:
                        ref_has_constraint = True

                if '$ref' in schema:
                    url, ref_sch = self.resolver.resolve(schema['$ref'])

                    ref_has_constraint = _schema_allows_no_undefined_props(ref_sch)
                if not ref_has_constraint and \
                   not (has_constraint or (schema.keys() & {'additionalProperties', 'unevaluatedProperties'})):
                    print(f"{self.filename}: {parent}: Missing additionalProperties/unevaluatedProperties constraint\n",
                          file=sys.stderr)

            for k, v in schema.items():
                self._check_schema_refs(v, parent=k, is_common=is_common,
                                        has_constraint=has_constraint)
        elif isinstance(schema, (list, tuple)):
            for i in range(len(schema)):
                self._check_schema_refs(schema[i], parent=parent, is_common=is_common,
                                        has_constraint=has_constraint)

    def check_schema_refs(self):
        id = self['$id'].rstrip('#')
        base1 = re.search('schemas/(.+)$', id)[1]
        base2 = self.filename.replace(self.filename[:-len(base1)], '')
        if not base1 == base2:
            print(f"{self.filename}: $id: Cannot determine base path from $id, relative path/filename doesn't match actual path or filename\n",
                  f"\t $id: {id}\n",
                  f"\tfile: {self.filename}",
                  file=sys.stderr)
            return

        self.paths = [
            (schema_base_url + 'schemas/', self.base_path),
            (schema_base_url + 'schemas/', schema_basedir + '/schemas/'),
        ]
        scope = self.validator.ID_OF(self)
        if scope:
            self.resolver.push_scope(scope)

        try:
            self._check_schema_refs(self)
        except jsonschema.RefResolutionError as exc:
            print(f"{self.filename}:\n{exc}", file=sys.stderr)



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-2-Clause
# Copyright 2018 Linaro Ltd.
# Copyright 2018 Arm Ltd.

import sys
import os
import argparse
import glob

import dtschema

verbose = False
show_unmatched = False
match_schema_file = None
compatible_match = False


class schema_group():
    def __init__(self, schema_file=""):
        if schema_file != "" and not os.path.exists(schema_file):
            exit(-1)

        self.validator = dtschema.DTValidator([schema_file])

    def check_node(self, tree, node, disabled, nodename, fullname, filename):
        # Hack to save some time validating examples
        if 'example-0' in node or 'example-' in nodename:
            return

        node['$nodename'] = [nodename]

        try:
            for error in self.validator.iter_errors(node, filter=match_schema_file,
                                                    compatible_match=compatible_match):

                # Disabled nodes might not have all the required
                # properties filled in, such as a regulator or a
                # GPIO meant to be filled at the DTS level on
                # boards using that particular node. Thus, if the
                # node is marked as disabled, let's just ignore
                # any error message reporting a missing property.
                if disabled or (isinstance(error.instance, dict) and
                   'status' in error.instance and
                   'disabled' in error.instance['status']):

                    if {'required', 'unevaluatedProperties'} & set(error.schema_path):
                        continue
                    elif error.context:
                        found = False
                        for e in error.context:
                            if {'required', 'unevaluatedProperties'} & set(e.schema_path):
                                found = True
                                break
                        if found:
                            continue

                if error.schema_file == 'generated-compatibles':
                    if not show_unmatched:
                        continue
                    print(f"{filename}: {fullname}: failed to match any schema with compatible: {node['compatible']}",
                          file=sys.stderr)
                    continue

                print(dtschema.format_error(filename, error, nodename=nodename, verbose=verbose),
                    file=sys.stderr)
        except RecursionError as e:
            print(os.path.basename(sys.argv[0]) + ": recursion error: Check for prior errors in a referenced schema", file=sys.stderr)

    def check_subtree(self, tree, subtree, disabled, nodename, fullname, filename):
        if nodename.startswith('__'):
            return

        try:
            disabled = ('disabled' in subtree['status'])
        except:
            pass

        self.check_node(tree, subtree, disabled, nodename, fullname, filename)
        if fullname != "/":
            fullname += "/"
        for name, value in subtree.items():
            if isinstance(value, dict):
                self.check_subtree(tree, value, disabled, name, fullname + name, filename)

    def check_dtb(self, filename):
        """Check the given DT against all schemas"""
        with open(filename, 'rb') as f:
            dt = self.validator.decode_dtb(f.read())
        for subtree in dt:
            self.check_subtree(dt, subtree, False, "/", "/", filename)


def main():
    global verbose
    global show_unmatched
    global match_schema_file
    global compatible_match

    ap = argparse.ArgumentParser(fromfile_prefix_chars='@',
        epilog='Arguments can also be passed in a file prefixed with a "@" character.')
    ap.add_argument("dtbs", nargs='*',
                    help="Filename or directory of devicetree DTB input file(s)")
    ap.add_argument('-s', '--schema', help="preparsed schema file or path to schema files")
    ap.add_argument('-p', '--preparse', help="preparsed schema file (deprecated, use '-s')")
    ap.add_argument('-l', '--limit', help="limit validation to schemas with $id matching LIMIT substring. " \
                    "Multiple substrings separated by ':' can be listed (e.g. foo:bar:baz).")
    ap.add_argument('-c', '--compatible-match', action="store_true",
                    help="limit validation to schema matching nodes' most specific compatible string")
    ap.add_argument('-m', '--show-unmatched',
        help="Print out node 'compatible' strings which don't match any schema.",
        action="store_true")
    ap.add_argument('-n', '--line-number', help="Obsolete", action="store_true")
    ap.add_argument('-v', '--verbose', help="verbose mode", action="store_true")
    ap.add_argument('-u', '--url-path', help="Additional search path for references (deprecated)")
    ap.add_argument('-V', '--version', help="Print version number",
                    action="version", version=dtschema.__version__)
    args = ap.parse_args()

    verbose = args.verbose
    show_unmatched = args.show_unmatched
    if args.limit:
        match_schema_file = args.limit.split(':')
    compatible_match = args.compatible_match

    # Maintain prior behaviour which accepted file paths by stripping the file path
    if args.url_path and args.limit:
        for i,match in enumerate(match_schema_file):
            for d in args.url_path.split(os.path.sep):
                if d and match.startswith(d):
                    match = match[(len(d) + 1):]
            match_schema_file[i] = match

    if args.preparse:
        sg = schema_group(args.preparse)
    elif args.schema:
        sg = schema_group(args.schema)
    else:
        sg = schema_group()

    for d in args.dtbs:
        if not os.path.isdir(d):
            continue
        for filename in glob.iglob(d + "/**/*.dtb", recursive=True):
            if verbose:
                print("Check:  " + filename)
            sg.check_dtb(filename)

    for filename in args.dtbs:
        if not os.path.isfile(filename):
            continue
        if verbose:
            print("Check:  " + filename)
        sg.check_dtb(filename)



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-2-Clause
# Copyright 2018 Linaro Ltd.
# Copyright 2018 Arm Ltd.


import os
import sys
import argparse
import glob

import ruamel.yaml
import dtschema

line_number = False
verbose = False


def check_doc(filename):
    ret = 0
    try:
        dtsch = dtschema.DTSchema(filename, line_numbers=line_number)
    except ruamel.yaml.YAMLError as exc:
        print(filename + ":" + str(exc.problem_mark.line + 1) + ":" +
              str(exc.problem_mark.column + 1) + ":", exc.problem, file=sys.stderr)
        return 1

    try:
        for error in sorted(dtsch.iter_errors(), key=lambda e: e.linecol):
            print(dtschema.format_error(filename, error, verbose=verbose), file=sys.stderr)
            ret = 1
    except:
        raise
        print(filename + ': error checking schema file', file=sys.stderr)
        return 1

    dtsch.check_schema_refs()

    return ret


def main():
    global verbose
    global line_number

    ap = argparse.ArgumentParser(fromfile_prefix_chars='@',
        epilog='Arguments can also be passed in a file prefixed with a "@" character.')
    ap.add_argument("yamldt", nargs='*', type=str,
                    help="Directory or filename of YAML encoded devicetree schema file")
    ap.add_argument('-v', '--verbose', help="verbose mode", action="store_true")
    ap.add_argument('-n', '--line-number', help="Print line and column numbers (slower)", action="store_true")
    ap.add_argument('-u', '--url-path', help="Additional search path for references")
    ap.add_argument('-V', '--version', help="Print version number",
                    action="version", version=dtschema.__version__)
    args = ap.parse_args()

    line_number = args.line_number
    verbose = args.verbose

    ret = 0
    for f in args.yamldt:
        if os.path.isdir(f):
            for filename in glob.iglob(f + "/**/*.yaml", recursive=True):
                ret |= check_doc(filename)
        else:
            ret |= check_doc(f)

    exit(ret)



#!/usr/bin/python3
# SPDX-License-Identifier: BSD-2-Clause
# Copyright 2022 Arm Ltd.

import argparse
import os
import sys
import pprint

import dtschema

strict = False


def main():
    ap = argparse.ArgumentParser()
    ap.add_argument("schema_files", type=str, nargs='*',
                    help="preparsed schema file or list of additional schema files/directories")
    ap.add_argument('-d', '--duplicates', help="Only output properties with more than one type", action="store_true")
    ap.add_argument('-v', '--verbose', help="Additional search path for references", action="store_true")
    ap.add_argument('-V', '--version', help="Print version number",
                    action="version", version=dtschema.__version__)
    args = ap.parse_args()

    dtval = dtschema.DTValidator(args.schema_files)
    props = dtval.property_get_all()

    if args.duplicates:
        tmp_props = {}
        for k, v in props.items():
            if len(v) > 1:
                tmp_props[k] = v

        props = tmp_props

    if args.verbose:
        prop_types = props
    else:
        prop_types = {}
        for k, v in props.items():
            prop_types[k] = []
            for l in v:
                if 'type' in l:
                    prop_types[k] += [l['type']]

    try:
        pprint.pprint(prop_types, compact=True)
        # flush output here to force SIGPIPE to be triggered
        # while inside this try block.
        sys.stdout.flush()
    except BrokenPipeError:
        # Python flushes standard streams on exit; redirect remaining output
        # to devnull to avoid another BrokenPipeError at shutdown
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())
        sys.exit(1)  # Python exits with error code 1 on EPIPE



#!/usr/bin/python3
# SPDX-License-Identifier: BSD-2-Clause
# Copyright 2022 Arm Ltd.

import argparse
import os
import sys
import pprint

import dtschema

strict = False

def main():
    ap = argparse.ArgumentParser()
    ap.add_argument("dtbfile", type=str, help="Schema directories and/or files")
    ap.add_argument('-s', '--schema', help="path to additional additional schema files")
    ap.add_argument('-V', '--version', help="Print version number",
                    action="version", version=dtschema.__version__)
    args = ap.parse_args()

    if not os.path.isfile(args.dtbfile):
        exit(1)

    if args.schema:
        schemas = [args.schema]
    else:
        schemas = []

    with open(args.dtbfile, 'rb') as f:
        dt = dtschema.DTValidator(schemas).decode_dtb(f.read())

    try:
        pprint.pprint(dt, compact=True)
        # flush output here to force SIGPIPE to be triggered
        # while inside this try block.
        sys.stdout.flush()
    except BrokenPipeError:
        # Python flushes standard streams on exit; redirect remaining output
        # to devnull to avoid another BrokenPipeError at shutdown
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())
        sys.exit(1)  # Python exits with error code 1 on EPIPE



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-2-Clause
# Copyright 2018 Linaro Ltd.
# Copyright 2019-2022 Arm Ltd.

import re
import sys
import argparse
import signal

import ruamel.yaml


def sigint_handler(signum, frame):
    sys.exit(-2)


signal.signal(signal.SIGINT, sigint_handler)


interrupt_template = """
        interrupt-parent = <&fake_intc{index}>;
        fake_intc{index}: fake-interrupt-controller {{
            interrupt-controller;
            #interrupt-cells = < {int_cells} >;
        }};
"""

example_template = """
    example-{example_num} {{
        #address-cells = <1>;
        #size-cells = <1>;

        {interrupt}

        {example}
    }};
}};
"""

example_header = """
/dts-v1/;
/plugin/; // silence any missing phandle references
"""
example_start = """
/{
    compatible = "foo";
    model = "foo";
    #address-cells = <1>;
    #size-cells = <1>;

"""

yaml = ruamel.yaml.YAML(typ='safe')


def main():
    ex = '// empty'
    ap = argparse.ArgumentParser()
    ap.add_argument("yamlfile", type=str,
                    help="Filename of YAML encoded schema input file")
    args = ap.parse_args()

    try:
        binding = yaml.load(open(args.yamlfile, encoding='utf-8').read())
        if not isinstance(binding, dict):
            exit(0)
    except ruamel.yaml.YAMLError as exc:
        print(args.yamlfile + ":" + str(exc.problem_mark.line + 1) + ":" +
              str(exc.problem_mark.column + 1) + ":", exc.problem, file=sys.stderr)
        exit(1)

    example_dts = example_header

    if 'examples' in binding.keys():
        for idx, ex in enumerate(binding['examples']):
            # Check if example contains a root node "/{"
            root_node = re.search(r'/\s*{', ex)

            if not root_node:
                try:
                    int_val = re.search(r'\sinterrupts\s*=\s*<([0-9a-zA-Z |()_]+)>', ex).group(1)
                    int_val = re.sub(r'\(.+|\)', r'0', int_val)
                    int_cells = len(int_val.strip().split())
                except:
                    int_cells = 0
                example_dts += example_start
                ex = '        '.join(ex.splitlines(True))
                if int_cells > 0:
                    int_props = interrupt_template.format(int_cells=int_cells, index=idx)
                else:
                    int_props = ""
                example_dts += example_template.format(example=ex, example_num=idx, interrupt=int_props)
            else:
                example_dts += ex
    else:
        example_dts += example_start
        example_dts += "\n};"

    print(example_dts)



# SPDX-License-Identifier: BSD-2-Clause
# Copyright 2021-2022 Arm Ltd.
# Python library for Devicetree schema validation

import sys
import struct

import libfdt
from libfdt import QUIET_NOTFOUND

import dtschema

u8 = struct.Struct('B')
s8 = struct.Struct('b')
u16 = struct.Struct('>H')
s16 = struct.Struct('>h')
u32 = struct.Struct('>L')
s32 = struct.Struct('>l')
u64 = struct.Struct('>Q')
s64 = struct.Struct('>q')

type_format = {
    'int8': s8,
    'uint8': u8,
    'int16': s16,
    'uint16': u16,
    'int32': s32,
    'uint32': u32,
    'int64': s64,
    'uint64': u64,
    'phandle': u32,
    'address': u32
}


def bytes_to_string(b):
    try:
        strings = b.decode(encoding='ascii').split('\0')

        count = len(strings) - 1
        if count > 0 and not len(strings[-1]):
            for string in strings[:-1]:
                if not string:
                    # While empty strings are valid, assume empty strings means
                    # not a string as we only get here when there are multiple
                    # types.
                    break
                if not string.isprintable():
                    break
            else:
                return strings[:-1]
    except:
        return None


def get_stride(prop_len, dim):

    # If multiple dimensions, check if one and only one dimension fits
    match = []
    for outer in range(dim[0][0], dim[0][1] + 1):
        for inner in range(dim[1][0], dim[1][1] + 1):
            if outer * inner == prop_len:
                match += [inner]

    if len(match) > 0:
        return match[0]

    if dim[1][0] > 0 and dim[1][0] == dim[1][1]:
        return dim[1][0]

    if dim[0][0] > 0 and dim[0][0] == dim[0][1]:
        return int(prop_len / dim[0][0])

    return prop_len

def prop_value(validator, nodename, p):
    # First, check for a boolean type
    if not len(p):
        return True

    dim = None
    data = bytes(p)
    fmt = None

    if nodename in {'__fixups__', 'aliases'}:
        return data[:-1].decode(encoding='ascii').split('\0')

    prop_types = set(validator.property_get_type(p.name))
    prop_types -= {'node'}

    # Filter out types impossible for the size of the property
    if len(prop_types) > 1:
        if len(p) % 8:
            prop_types -= {'int64', 'uint64', 'int64-array', 'uint64-array'}
        if len(p) % 4:
            prop_types -= {'int32', 'uint32', 'int32-array', 'uint32-array', 'phandle', 'phandle-array'}
        if len(p) % 2:
            prop_types -= {'int16', 'uint16', 'int16-array', 'uint16-array'}

        if len(p) > 4:
            prop_types -= {'int32', 'uint32', 'phandle'}
        else:
            prop_types -= {'int64', 'uint64', 'int64-array', 'uint64-array'}
        if len(p) > 2:
            prop_types -= {'int16', 'uint16'}
        else:
            prop_types -= {'int32', 'uint32', 'int32-array', 'uint32-array', 'phandle', 'phandle-array'}
        if len(p) > 1:
            prop_types -= {'int8', 'uint8'}
        else:
            prop_types -= {'int16', 'uint16', 'int16-array', 'uint16-array'}
        if len(p) > 0:
            prop_types -= {'flag'}

        if prop_types >= {'phandle', 'phandle-array'}:
            prop_types -= {'phandle'}

        # Drop the unsigned type if both signed and unsigned type exists
        if prop_types >= {'int64', 'uint64'}:
            prop_types -= {'uint64'}
        if prop_types >= {'int32', 'uint32'}:
            prop_types -= {'uint32'}
        if prop_types >= {'int16', 'uint16'}:
            prop_types -= {'uint16'}
        if prop_types >= {'int8', 'uint8'}:
            prop_types -= {'uint8'}

    if len(prop_types) > 1:
        if {'string', 'string-array'} & prop_types:
            str = bytes_to_string(data)
            if str:
                return str
            # Assuming only one other type
            try:
                fmt = prop_types.difference({'string', 'string-array'}).pop()
            except:
                return data
        else:
            #print(p.name + ': multiple types found', file=sys.stderr)
            # HACK around a type collision. Since 4 bytes could be either type,
            # let's hope a value of 1-4 is not a phandle
            if p.name == "dma-masters":
                if len(p) == 4 and (0 < type_format['uint32'].unpack(data)[0] <= 4):
                    fmt = 'uint32'
                else:
                    fmt = 'phandle-array'
            else:
                fmt = None
    elif len(prop_types) == 1:
        fmt = prop_types.pop()

    if not fmt:
        # Primarily for aliases properties
        try:
            s = data.decode(encoding='ascii')
            if s.endswith('\0'):
                s = s[:-1]
                if s.isprintable():
                    return [s]
        except:
            pass
        if not len(data) % 4:
            fmt = 'uint32-array'
        else:
            #print(p.name + ': no type found', file=sys.stderr)
            return data

    if fmt.startswith('string'):
        if not data.endswith(b'\0'):
            # Wrong data type?, skip decoding to force errors
            return data
        return data[:-1].decode(encoding='ascii').split('\0')

    if {'flag'} & prop_types:
        if len(data):
            if fmt == 'flag':
                print('{prop}: boolean property with value {val}'.format(prop=p.name, val=data),
                      file=sys.stderr)
                return data
        else:
            return True

    val_int = list()
    #print(p.name, fmt,  bytes(p))
    try:
        type_struct = type_format[fmt.split('-', 1)[0]]
        for i in type_struct.iter_unpack(data):
            val_int += [dtschema.sized_int(i[0], size=(type_struct.size * 8))]
    except:
        print('{prop}: size ({len}) error for type {fmt}'.format(prop=p.name, len=len(p), fmt=fmt), file=sys.stderr)
        if len(p) == 4:
            type_struct = type_format['uint32']
        elif len(p) == 2:
            type_struct = type_format['uint16']
        elif len(p) == 1:
            type_struct = type_format['uint8']
        else:
            return data
        for i in type_struct.iter_unpack(data):
            val_int += [dtschema.sized_int(i[0], size=(type_struct.size * 8))]

    if 'matrix' in fmt or fmt in {'phandle', 'phandle-array', 'address'}:
        dim = validator.property_get_type_dim(p.name)
        if dim:
            stride = get_stride(len(val_int), dim)

            #print(p.name, dim, stride, len(val_int))
            return [val_int[i:i+stride] for i in range(0, len(val_int), stride)]
        else:
            return [val_int]
    elif 'array' not in fmt and len(val_int) == 1:
        return val_int[0]

    return val_int


def node_props(validator, fdt, nodename, offset):
    props_dict = {}
    poffset = fdt.first_property_offset(offset, QUIET_NOTFOUND)
    while poffset >= 0:
        p = fdt.get_property_by_offset(poffset)
        props_dict[p.name] = prop_value(validator, nodename, p)

        poffset = fdt.next_property_offset(poffset, QUIET_NOTFOUND)

    return props_dict


phandles = {}
phandle_loc = []


def process_fixups(validator, fdt, nodename, offset):
    if nodename != '__fixups__':
        return
    props = node_props(validator, fdt, nodename, offset)
    global phandle_loc
    phandle_loc += [s for l in props.values() for s in l]


def process_local_fixups(fdt, nodename, path, offset):
    global phandle_loc

    if nodename:
        path += '/' + nodename

    poffset = fdt.first_property_offset(offset, QUIET_NOTFOUND)
    while poffset >= 0:
        p = fdt.get_property_by_offset(poffset)

        for i in type_format['uint32'].iter_unpack(bytes(p)):
            phandle_loc += [path + ':' + p.name + ':' + str(i[0])]

        poffset = fdt.next_property_offset(poffset, QUIET_NOTFOUND)

    offset = fdt.first_subnode(offset, QUIET_NOTFOUND)
    while offset >= 0:
        nodename = fdt.get_name(offset)
        process_local_fixups(fdt, nodename, path, offset)

        offset = fdt.next_subnode(offset, QUIET_NOTFOUND)


def fdt_scan_node(validator, fdt, nodename, offset):
    if nodename == '__fixups__':
        process_fixups(validator, fdt, nodename, offset)
        return
    if nodename == '__local_fixups__':
        process_local_fixups(fdt, '', '', offset)
        return
    if nodename.startswith('__'):
        return

    node_dict = node_props(validator, fdt, nodename, offset)
    if 'phandle' in node_dict:
        #print('phandle', node_dict['phandle'])
        phandles[node_dict['phandle']] = node_dict

    offset = fdt.first_subnode(offset, QUIET_NOTFOUND)
    while offset >= 0:
        nodename = fdt.get_name(offset)
        node = fdt_scan_node(validator, fdt, nodename, offset)
        if node is not None:
            node_dict[nodename] = node

        offset = fdt.next_subnode(offset, QUIET_NOTFOUND)

    return node_dict


phandle_args = {
    # phandle+args properties with fixed arg size or which don't match standard
    # 'foos' and '#foo-cells' pattern
    'assigned-clocks': '#clock-cells',
    'assigned-clock-parents': '#clock-cells',
    'cooling-device': '#cooling-cells',
    'interrupts-extended': '#interrupt-cells',
    'interconnects': '#interconnect-cells',
    'mboxes': '#mbox-cells',
    'sound-dai': '#sound-dai-cells',
    'msi-parent': '#msi-cells',
    'msi-ranges': '#interrupt-cells',
    'gpio-ranges': 3,

    'memory-region': None,
}


def _get_cells_size(node, cellname):
    if isinstance(cellname, int):
        return cellname
    elif cellname in node:
        return node[cellname]
    else:
        return 0


def _check_is_phandle(prop_path, cell):
    path = prop_path + ':' + str(cell * 4)
    return path in phandle_loc


def _get_phandle_arg_size(prop_path, idx, cells, cellname):
    if not cells:
        return 0
    phandle = cells[0]
    if phandle == 0 or not cellname:
        return 1
    if phandle == 0xffffffff:
        # Use fixups data if available (examples)
        # Mixing unresolved and resolved phandles doesn't work
        if _check_is_phandle(prop_path, idx):
            cell_count = 1
            while (cell_count < len(cells)) and not _check_is_phandle(prop_path, idx + cell_count):
                cell_count += 1

            return cell_count
        else:
            return 0

    if phandle not in phandles:
        return 0

    node = phandles[phandle]

    return _get_cells_size(node, cellname) + 1


def fixup_phandles(validator, dt, path=''):
    for k, v in dt.items():
        if isinstance(v, dict):
            fixup_phandles(validator, v, path=path + '/' + k)
            continue
        elif not {'phandle-array'} & set(validator.property_get_type(k)):
            continue
        elif validator.property_has_fixed_dimensions(k):
            continue
        elif not isinstance(v, list) or (len(v) > 1 or not isinstance(v[0], list)):
            # Not a matrix or already split, nothing to do
            continue
        elif k in phandle_args:
            cellname = phandle_args[k]
        elif k.endswith('s') and 'gpio' not in k:
            cellname = '#' + k[:-1] + '-cells'
            #print(k, v)
            i = _get_phandle_arg_size(path + ':' + k, 0, v[0], cellname)
            if i == 0:
                continue
        else:
            continue

        i = 0
        dt[k] = []
        val = v[0]
        #print(k, v)
        while i < len(val):
            #print(k, v, file=sys.stderr)
            cells = _get_phandle_arg_size(path + ':' + k, i, val[i:], cellname)
            if cells == 0:
                #print(k, v)
                break

            # Special case for msi-ranges which has an additional span cell
            if k == 'msi-ranges':
                cells += 1

            # Special case for interconnects which is pairs of phandles+args
            if k == 'interconnects':
                cells += _get_phandle_arg_size(path + ':' + k, i + cells, val[i + cells:], cellname)

            dt[k] += [val[i:i + cells]]
            #print(k, dt[k], file=sys.stderr)

            i += cells


def fixup_gpios(dt):
    if 'gpio-hog' in dt:
        return
    for k, v in dt.items():
        if isinstance(v, dict):
            fixup_gpios(v)
        elif (k.endswith('-gpios') or k.endswith('-gpio') or k in {'gpio', 'gpios'}) and \
             not k.endswith(',nr-gpios') and \
             isinstance(v, list):
            i = 0
            dt[k] = []
            val = v[0]
            while i < len(val):
                phandle = val[i]
                if phandle == 0:
                    cells = 0
                elif phandle == 0xffffffff:
                    #print(val, file=sys.stderr)
                    try:
                        cells = val.index(0xffffffff, i + 1, -1)
                    except:
                        cells = len(val)
                    #print(cells, file=sys.stderr)
                    cells -= (i + 1)
                else:
                    #print(k, v, file=sys.stderr)
                    node = phandles[phandle]
                    cells = _get_cells_size(node, '#gpio-cells')

                dt[k] += [val[i:i + cells + 1]]

                i += cells + 1


def fixup_interrupts(dt, icells):
    if 'interrupt-parent' in dt and isinstance(dt['interrupt-parent'], list):
        phandle = dt['interrupt-parent'][0][0]
        if phandle == 0xffffffff:
            del dt['interrupt-parent']
        else:
            icells = _get_cells_size(phandles[phandle], '#interrupt-cells')

    for k, v in dt.items():
        if isinstance(v, dict):
            if '#interrupt-cells' in dt:
                icells = _get_cells_size(dt, '#interrupt-cells')
            fixup_interrupts(v, icells)
        elif k == 'interrupts' and not isinstance(v, bytes):
            i = 0
            dt[k] = []
            val = v[0]
            while i < len(val):
                dt[k] += [val[i:i + icells]]
                i += icells
        elif k == 'interrupt-map' and not isinstance(v, bytes):
            icells = _get_cells_size(dt, '#interrupt-cells')
            ac = _get_cells_size(dt, '#address-cells')
            i = 0
            dt[k] = []
            val = v[0]
            phandle = val[ac + icells]
            if phandle == 0xffffffff:
                # Assume uniform sizes (same interrupt provider)
                try:
                    cells = val.index(0xffffffff, ac + icells + 1) - (ac + icells)
                except ValueError:
                    cells = len(val)
                while i < len(val):
                    dt[k] += [val[i:i + cells]]
                    i += cells
            else:
                while i < len(val):
                    p_icells = _get_cells_size(phandles[phandle], '#interrupt-cells')
                    if '#address-cells' in phandles[phandle]:
                        p_ac = _get_cells_size(phandles[phandle], '#address-cells')
                    else:
                        p_ac = 0

                    cells = ac + icells + 1 + p_ac + p_icells
                    dt[k] += [val[i:i + cells]]
                    i += cells


def fixup_addresses(validator, dt, ac, sc):
    for k, v in dt.items():
        if isinstance(v, dict):
            if '#address-cells' in dt:
                ac = _get_cells_size(dt,'#address-cells')
            if '#size-cells' in dt:
                sc = _get_cells_size(dt, '#size-cells')
            fixup_addresses(validator, v, ac, sc)
        elif 'address' in validator.property_get_type(k):
            i = 0
            dt[k] = []
            val = v[0]
            while i < len(val):
                dt[k] += [val[i:i + ac + sc]]
                i += ac + sc
        elif k in {'ranges', 'dma-ranges'} and not isinstance(v, bool):
            child_cells = _get_cells_size(dt, '#address-cells')
            child_cells += _get_cells_size(dt, '#size-cells')
            i = 0
            dt[k] = []
            val = v[0]
            while i < len(val):
                dt[k] += [val[i:i + ac + child_cells]]
                i += ac + child_cells


def fdt_unflatten(validator, dtb):
    fdt = libfdt.Fdt(dtb)

    offset = fdt.first_subnode(-1, QUIET_NOTFOUND)
    dt = fdt_scan_node(validator, fdt, '/', offset)

    #print(phandle_loc)
    fixup_gpios(dt)
    fixup_interrupts(dt, 1)
    fixup_addresses(validator, dt, 2, 1)
    fixup_phandles(validator, dt)

#    pprint.pprint(dt, compact=True)
    return dt



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-2-Clause
# Copyright 2022 Arm Ltd.

import os
import argparse

import dtschema


def main():
    ap = argparse.ArgumentParser(fromfile_prefix_chars='@',
        epilog='Arguments can also be passed in a file prefixed with a "@" character.')
    ap.add_argument("compatible_str", nargs='+',
                    help="1 or more compatible strings to check for a match")
    ap.add_argument('-q', '--quiet', action="store_true",
                    help="Suppress printing matches")
    ap.add_argument('-v', '--invert-match', action="store_true",
                    help="invert sense of matching, printing compatible which don't match")
    ap.add_argument('-V', '--version', help="Print version number",
                    action="version", version=dtschema.__version__)
    ap.add_argument('-s', '--schema', required=True, help="path to processed schema file or schema directory")
    args = ap.parse_args()

    if args.schema != "" and not os.path.exists(args.schema):
        exit(-1)

    undoc_compats = dtschema.DTValidator([args.schema]).get_undocumented_compatibles(args.compatible_str)

    if args.invert_match:
        if len(undoc_compats) > 0:
            if not args.quiet:
                print(*undoc_compats, sep="\n")
            return 0
    else:
        matches = set(args.compatible_str).difference(undoc_compats)
        if len(matches) > 0:
            if not args.quiet:
                print(*matches, sep="\n")
            return 0

    return -1



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-2-Clause
# Copyright 2023-2024 Arm Ltd.

import sys
import argparse
import urllib

import dtschema


def path_list_to_str(path):
    return '/' + '/'.join(path)


def prop_generator(schema, path=[]):
    if not isinstance(schema, dict):
        return
    for prop_key in ['properties', 'patternProperties']:
        if prop_key in schema:
            for p, sch in schema[prop_key].items():
                yield path + [prop_key, p], sch
                yield from prop_generator(sch, path=path + [prop_key, p])


def _ref_to_id(schema_id, ref):
    ref = urllib.parse.urljoin(schema_id, ref)
    if '#/' not in ref:
        ref += '#'
    return ref


def _prop_in_schema(prop, schema, schemas):
    for p, sch in prop_generator(schema):
        if p[1] == prop:
            return True

    if 'allOf' in schema:
        for e in schema['allOf']:
            if '$ref' in e:
                ref_id = _ref_to_id(schema['$id'], e['$ref'])
                if ref_id in schemas:
                    if _prop_in_schema(prop, schemas[ref_id], schemas):
                        return True

    if '$ref' in schema:
        ref_id = _ref_to_id(schema['$id'], schema['$ref'])
        if ref_id in schemas and _prop_in_schema(prop, schemas[ref_id], schemas):
            return True

    return False


def check_removed_property(schema_id, base, schemas):
    for p, sch in prop_generator(base):
        if not _prop_in_schema(p[1], schemas[schema_id], schemas):
            print(f'{schema_id}{path_list_to_str(p)}: existing property removed', file=sys.stderr)


def check_deprecated_property(schema_id, base, schemas):
    for p, sch in prop_generator(base):
        if isinstance(sch, dict) and 'deprecated' in sch:
            continue
        schema = schema_get_from_path(schemas[schema_id], p)
        if schema and isinstance(schema, dict) and 'deprecated' in schema:
            print(f'{schema_id}{path_list_to_str(p)}: existing property deprecated', file=sys.stderr)


def schema_get_from_path(sch, path):
    for p in path:
        try:
            sch = sch[p]
        except:
            return None
    return sch


def check_new_items(schema_id, base, new):
    for p, sch in prop_generator(new):
        if not isinstance(sch, dict) or 'minItems' not in sch:
            continue

        new_min = sch['minItems']
        base_min = schema_get_from_path(base, p + ['minItems'])

        if base_min and new_min > base_min:
            print(f'{schema_id}{path_list_to_str(p)}: required entries increased from {base_min} to {new_min}', file=sys.stderr)


def _get_required(schema):
    required = []
    for k in {'allOf', 'oneOf', 'anyOf'} & schema.keys():
        for sch in schema[k]:
            if 'required' not in sch:
                continue
            required += sch['required']

    if 'required' in schema:
        required += schema['required']

    return set(required)


def _check_required(schema_id, base, new, path=[]):
    if not isinstance(base, dict) or not isinstance(new, dict):
        return

    base_req = _get_required(base)
    new_req = _get_required(new)

    if not new_req:
        return

    diff = new_req - base_req
    if diff:
        print(f'{schema_id}{path_list_to_str(path)}: new required properties added: {", ".join(diff)}', file=sys.stderr)
        return


def check_required(schema_id, base, new):
    _check_required(schema_id, base, new)

    for p, sch in prop_generator(new):
        _check_required(schema_id, schema_get_from_path(base, p), sch, path=p)


def main():
    ap = argparse.ArgumentParser(description="Compare 2 sets of schemas for possible ABI differences")
    ap.add_argument("baseline", type=str,
                    help="Baseline schema directory or preprocessed schema file")
    ap.add_argument("new", type=str,
                    help="New schema directory or preprocessed schema file")
    ap.add_argument('-V', '--version', help="Print version number",
                    action="version", version=dtschema.__version__)
    args = ap.parse_args()

    base_schemas = dtschema.DTValidator([args.baseline]).schemas
    schemas = dtschema.DTValidator([args.new]).schemas

    if not schemas or not base_schemas:
        return -1

    for schema_id, sch in schemas.items():
        if schema_id not in base_schemas or 'generated' in schema_id:
            continue

        check_required(schema_id, base_schemas[schema_id], sch)
        check_removed_property(schema_id, base_schemas[schema_id], schemas)
        check_deprecated_property(schema_id, base_schemas[schema_id], schemas)
        check_new_items(schema_id, base_schemas[schema_id], sch)
